# run "FUNC" with gene expression data from Allen Brain Atlas 
# over flexible expression cutoff
# wilcoxon rank test or hypergeometric test
# with adult microarray data, developmental RNA-seq (5 age categories) or derived developmental effect score


###############
# MAIN function
###############

### Arguments
# genes: vector of 0/1 (hypergeometric) or float (wilcoxon) with gene identifiers as names (ensembl, entrez or hgnc), or chromosomal regions chr:start-stop (only for hypergeometric)
# dataset: "adult", "5_stages" or "dev_effect"
# test: "hyper" or "wilcoxon"
# cutoff_quantiles: numeric values in ]0,1[
# n_randsets: number of random-sets
# gene_len: randomset is dependent on length of genes
# circ_chrom: for regions input: random regions are on same chrom and allowed to overlap multiple bg-regions
# ref_genome: "grch37" or "grch38" for region input and gen_len=T

#######################################################################################
# 1. check arguments and define parameters
# 2. load expression data
# 3. loop over age categories and cutoffs
#	3a) create input
#	3b) run func
#	3c) summarize func output
# 4. create output
#######################################################################################

aba_enrich=function(genes,dataset="adult",test="hyper",cutoff_quantiles=seq(0.1,0.9,0.1),n_randsets=1000, gene_len=FALSE, circ_chrom=FALSE, ref_genome="grch37"){


	#####	1. Check arguments and define parameters
		
	## Define dataset dependent parameters
	if (dataset=="adult"){
		folder_ext="adult"
		root_node="Allen:4005"
	} else if (dataset=="5_stages"){
		folder_ext="developmental"
		root_node="Allen:10153"
	} else if (dataset=="dev_effect"){
		folder_ext="developmental"
		root_node="Allen:10153"
	} else {
		stop("Not a valid dataset. Please use 'adult', '5_stages' or 'dev_effect'.")
	}
	
	## Check arguments
	# general
	message("Checking arguments...")
	if (length(genes)==0){
		stop("Please enter genes.")
	}	
	if (length(names(genes))==0){
		stop("Please add gene identifiers as names to 'genes' vector.")
	}
	if (class(cutoff_quantiles)!="numeric"){
		stop("Please enter numeric cutoff_quantiles.")
	}	
	if (min(cutoff_quantiles)<=0 | max(cutoff_quantiles)>=1){
		stop("Please enter cutoff_quantiles between 0 and 1 (exclusive).")
	}			
	valid_datasets = c("adult","5_stages","dev_effect")
	if (!(dataset %in% valid_datasets)){
		stop("Not a valid dataset. Please use 'adult', '5_stages' or 'dev_effect'.")	
	}
	if (length(n_randsets)>1 || !is.numeric(n_randsets) || n_randsets<1){
		stop("Please define 'n_randsets' as a positive integer.")
	}
	if (n_randsets != round(n_randsets)){
		n_randsets = round(n_randsets)
		warning(paste("'n_randsets' is expected to be an integer and was rounded to ",n_randsets,".",sep=""))
	}
	if (!is.logical(gene_len)){
		stop("Please set gene_len to TRUE or FALSE.")
	}
	if (!is.logical(circ_chrom)){
		stop("Please set circ_chrom to TRUE or FALSE.")
	}
	if (!(ref_genome %in% c("grch37", "grch38"))){
		stop("Not a valid ref_genome. Please use 'grch37' or 'grch38'.")
	}
	# test-specific arguments
	if (test=="hyper"){
		if (!all(genes %in% c(0,1))){
			stop("Not a valid 'genes' argument for hypergeometric test. Please use a vector of 0/1.")	
		}	
		if (sum(genes)==0){
			stop("Only 0 in genes vector. Please enter test genes.")	
		}
	} else	if (test=="wilcoxon"){
		if (!is.numeric(genes)){
			stop("Not a valid 'genes' argument. Please use a numeric vector.")	
		}
		if (gene_len){
			warning("Unused argument: 'gene_len = TRUE'.")
		}
		if(length(genes) < 2){
			stop("Only one gene provided as input.")
		}
	} else (stop("Not a valid test. Please use 'hyper' or 'wilcoxon'."))
	
	# Create input and output folder
	directory = tempdir()
	
	# load gene_list and get gene identifier
	gene_symbols = get(paste("gene_symbols",folder_ext,sep="_"))	
	blocks = FALSE
	identifier = detect_identifier(names(genes)[1])	
	
	# load gene_coords
	gene_coords = get(paste("gene_coords_", ref_genome, sep=""))
	
	if (identifier == "blocks"){
		identifier = "hgnc_symbol" # gene-name 
		blocks = TRUE
		# check that background region is specified
		if (sum(genes) == length(genes)){
			stop("All values of the 'genes' input are 1. Using chromosomal regions as input requires defining background regions with 0.")
		}
		# check that test is hyper
		if (test != "hyper"){
			stop("chromosomal regions can only be used with test='hyper'.")
		}	
		# warn if gene_len=TRUE, although regions are used
		if (gene_len){
			warning("Unused argument: 'gene_len = TRUE'.")
		}	
		
		# convert coords from genes-vector to bed format, SORT, and extract genes overlapping test regions
		regions = get_genes_from_regions(genes, gene_coords, circ_chrom) # gene_coords from sysdata.rda HGNC
		test_regions = regions[[1]]		
		bg_regions = regions[[2]]
		genes = regions[[3]]

		# avoid scientific notation in regions (read in c++)
		test_regions = format(test_regions, scientific=FALSE, trim=TRUE)
		bg_regions = format(bg_regions, scientific=FALSE, trim=TRUE)

		message("Candidate regions:")
		print(test_regions)
		message("Background regions:")
		print(bg_regions)		
	
		# write regions to files
		write.table(test_regions,file=paste(directory, "/test_regions.bed",sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE,sep="\t")
		write.table(bg_regions,file=paste(directory, "/bg_regions.bed",sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE,sep="\t")		
	} else {
		if (circ_chrom == TRUE){
			# warn if circ_chrom=TRUE, although individual genes are used
			warning("Unused argument: 'circ_chrom = TRUE'.")
		}
		# NEW: check for multiple assignment for one gene (add to line 160)
		genetab = unique(data.frame(genes, names(genes))) # allow for multiple assignment of the same value
		multi_genes = sort(unique(genetab[,2][duplicated(genetab[,2])]))
		if (length(multi_genes) > 0){
			stop(paste("Genes with multiple assignment in input:", paste(multi_genes,collapse=", ")))
		}
	}
	if (ref_genome == "grch38" & !(blocks | gene_len)){
		# warn if ref-genome is switched to hg20 but not used
		warning("Unused argument: 'ref_genome = grch38'.")
	}

	gene_list = gene_symbols[,identifier]	

	# restrict to genes that have expression data annotated and warn about the rest
	remaining_genes = genes[names(genes) %in% gene_list] # restrict
	not_in = unique(names(genes)[!(names(genes) %in% names(remaining_genes))]) # removed
	if (length(not_in) > 0 && !blocks){
		not_in_string = paste(not_in,collapse=", ")
		warning(paste("No expression data for genes: ",not_in_string,".\n  These genes were not included in the analysis.",sep=""))
	}
	# restrict to genes that have coordinates and warn about the rest
	if (gene_len){
		no_coords = unique(names(remaining_genes)[!(names(remaining_genes) %in% gene_coords[,identifier])]) # removed
		if (length(no_coords) > 0){
			remaining_genes = remaining_genes[!(names(remaining_genes) %in% no_coords)] # restrict
			no_coords_string = paste(no_coords,collapse=", ")
			warning(paste("No coordinates available for genes: ",no_coords_string,".\n  These genes were not included in the analysis.",sep=""))
			not_in = c(not_in, no_coords)
		}
	}
	# after removing genes without expression data or coordinates: are enough genes left?
	if (length(not_in) > 0){
		# is any gene in the data? if not, stop.
		if (length(remaining_genes) == 0) {
			stop("No requested genes in data.")
		}
		# for 0/1 data: are test genes and background genes in the data (background only if background specified)?
		if (test=="hyper" & sum(remaining_genes)==0) {
			stop("No requested test genes in data.")
		}
		if (test=="hyper" & sum(genes)!=length(genes) & sum(remaining_genes)==length(remaining_genes)) {
			stop("No requested background genes in data.")
		}
	}
	# candidate genes	
	if (test=="hyper"){					
		interesting_genes = names(remaining_genes)[remaining_genes==1] 
	}

	#####	2. Load expression data
	
	# load pre input
	message("Loading dataset...")
	pre_input = load_by_name(paste("dataset",dataset,sep="_"))
	
	# TODO: save cutoff quantiles for all datasets for default cutoff_quantile values.
		
	 # compute cutoffs (tapply lasts much longer - use only when needed)
	message("Computing cutoffs... ")
	cutoff_quantiles = sort(unique(cutoff_quantiles))
	if (dataset=="5_stages"){
		cutoff_list = tapply(pre_input$signal, pre_input$age_category, function(x) quantile(x, probs=cutoff_quantiles),simplify=FALSE)
	} else {	# this if-else is only to save time, tapply would work in both cases
		cutoff_list = list()
		cutoff_list[[1]] = quantile(pre_input$signal,probs=cutoff_quantiles)
		names(cutoff_list) = pre_input$age_category[1]
	}

	# select gene identifier
	pre_input = pre_input[,c("age_category",identifier,"structure","signal")]
	colnames(pre_input)[2] = "gene_id"

	# restrain input to required genes (unless test=hypergeometric and background genes are not defined - all other genes are background in this case)
	# else exlude NAs if entrez-ID, (hgnc and ensembl dont have NAs)
	if (!(test=="hyper" & sum(genes) == length(genes))){		
		message("Select requested genes...")
		pre_input = pre_input[pre_input[,2] %in% names(remaining_genes),]
	} else if (identifier == "entrezgene"){
		message("Exclude NAs...")	
		pre_input = pre_input[!is.na(pre_input$gene_id),]
	}			
	# aggregate expression per gene (duplicates due to different identifiers)
	message("Checking for gene duplicates to aggregate... ")
	# simple aggregate would take very long for large input vectors
	# number of structure-age-combinations: if all structures are present in all ages they can simply be multiplied -> thats the case
	# n=nrow(unique(pre_input[,c("age_category","structure")]))
	n = length(unique(pre_input$age_category))*length(unique(pre_input$structure))
	ngenes = table(pre_input$gene_id)
	bose = ngenes[ngenes>n]
	if (length(bose)>0){
		message(" Aggregate expression per gene...")
		part1 = pre_input[pre_input$gene_id %in% names(bose),]
		part2 = pre_input[!(pre_input$gene_id %in% names(bose)),]
		part1 = aggregate(part1$signal,by=list("age_category"=part1$age_category,"gene_id"=part1$gene_id,"structure"=part1$structure),mean)
		colnames(part1)[4]="signal"
		pre_input = rbind(part1,part2)
		message(" Done.")
	} else {message(" No gene duplicates - no aggregation needed.")}

	# Add "Allen:"-string to brain region IDs
	pre_input$structure = paste("Allen:", pre_input$structure, sep="")
	
	# load ontology and write files to tmp
	term = get(paste("term",folder_ext,sep="_"))
	term2term = get(paste("term2term",folder_ext,sep="_"))
	graph_path = get(paste("graph_path",folder_ext,sep="_"))
	write.table(term,file=paste(directory, "/term.txt",sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE,sep="\t")
	write.table(term2term,file=paste(directory, "/term2term.txt",sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE,sep="\t")
	write.table(graph_path,file=paste(directory, "/graph_path.txt",sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE,sep="\t")

	#####	3. loop over age categories and cutoffs
	# initialize output
	out = data.frame()
	for (s in unique(pre_input$age_category)){
		stage_input = pre_input[pre_input$age_category==s,2:4]	
		# get cutoffs		
		cutoff = cutoff_list[[as.character(s)]]
		colnames(stage_input)[3] = "signal"		
		for (i in 1:length(cutoff)){
			
			##	3a) create input

			if (dataset=="5_stages"){
				message(paste("Preparing age category ", s," with ",names(cutoff[i]),"-quantile expression cutoff for FUNC...",sep=""))			
			} else {
				message(paste("Preparing data with ",names(cutoff[i]),"-quantile expression cutoff for FUNC...",sep=""))	
			}
			# restrain input with cutoff
			message(" Apply cutoff...")
			input = stage_input[stage_input$signal >= cutoff[i],]		
			
			# for Hypergeometric Test: 0 for all genes, then convert to 1 for interesting genes (merge not possible if no background genes are defined)
			message(" Check that there are sufficient genes above cutoff...")
			breaky = FALSE
			if (test=="hyper"){	
				# do input data have both 0 and 1? else break (FUNC would throw error and summary-file would not be generated)
				# now that cutoffs get computed for all genes (not just input) both could be missing
				if (nrow(input)==0){ 
					message(paste("  Warning: No statistics for cutoff >= ",cutoff_quantiles[i],". No input gene expression above cutoff.",sep=""))
					breaky = TRUE
				} else {	
					input$signal = 0
					input[input[,1] %in% interesting_genes,"signal"] = 1
					if (sum(input$signal)==0){
						message(paste("  Warning: No statistics for cutoff >= ",cutoff_quantiles[i],". No test gene expression above cutoff.",sep=""))
						breaky = TRUE
					} else if (sum(input$signal)==nrow(input)){
						message(paste("  Warning: No statistics for cutoff >= ",cutoff_quantiles[i],". No background gene expression above cutoff.",sep=""))
						breaky = TRUE
					}
				}		
			# for Wilcoxon test: replace expression signal with score
			} else	if (test=="wilcoxon"){
				if (length(unique(input[,1]))<2){
					message(paste("  Warning: No statistics for cutoff >= ",cutoff_quantiles[i],". Less than 2 genes show expression above cutoff.",sep=""))
					breaky = TRUE
				}
				if (!breaky){
					score = genes[match(input$gene_id,names(genes))]
					input$signal = score
				}		
			}	
			# if first cutoff fails, no output produced, else return output from previous cutoffs
			if (breaky){
				if (i==1){
					stop ("Expression cutoffs too high.")
				} else {
					break
				}	
			}
			message(" Rearrange input for FUNC...")
			# prepare input data (infile-data and root like in separate_taxonomies.pl)
			
			# "infile-data": genes and associated scores (wilcox) 
			# "Allen:4005-changed": one column with test genes (hyper)
			if (test=="hyper"){
				# subset to test genes
				infile_data = data.frame(genes=unique(input[input[,3]==1,1]))
			} else if (test=="wilcoxon"){
				infile_data = unique(input[,c(1,3)])
			}
	
			# create "root" dataframe, add genomic positions if block option is used
			# add gene-coordinates, despite for classic FUNC option (user defined genes might get lost because they have no annotated position - there are some from ABA which are not in gene_coords)			
			if (blocks || gene_len){
				# each row contains gene and a string with all annotaded brain regions
				xx=tapply(input[,2],input[,1],function(x){paste(x,collapse=" ")}) # paste annotations
				gene = as.character(names(xx))
				# add coordinates
				gene_position = gene_coords[match(gene, gene_coords[,identifier]),1:3]
				root = data.frame(genes=gene, gene_position ,goterms=as.character(xx))
				# remove genes with unknown coordinates (possible for default bg in gene_len-option) 
				if (gene_len){					
					root = root[!is.na(root[,3]),] 			
				}
				# NEW just in case - avoid scientific notation of gene coordinates	
				root[,3:4] = format(root[,3:4], scientific=FALSE, trim=TRUE)				
			} else {
				# each row contains gene and a string with all annotaded brain regions
				xx = tapply(input[,2],input[,1],function(x){paste(x,collapse=" ")}) # paste annotations
				root = data.frame(genes=as.character(names(xx)),goterms=as.character(xx))
			}		
			# write input-files to tmp-directory
			# last priority rename root and infile-data files to all_genes_annotation and test_genes
			write.table(infile_data,sep="\t",quote=FALSE,col.names=FALSE,row.names=FALSE,file=paste(directory,"/infile-data",sep=""))
			write.table(root,sep="\t",quote=FALSE,col.names=FALSE,row.names=FALSE,file=paste(directory,"/",root_node,sep=""))

			###	3b) run func
			if (test=="hyper"){
				# randomset
				if (blocks & circ_chrom){
					hyper_randset(paste(directory,"/",root_node,sep=""), n_randsets, directory, root_node, "roll")
				} else if (blocks){
					hyper_randset(paste(directory,"/",root_node,sep=""), n_randsets, directory, root_node,"block")				
				} else if (gene_len){						
					hyper_randset(paste(directory,"/",root_node,sep=""), n_randsets, directory, root_node, "gene_len")	
				} else {						
					hyper_randset(paste(directory,"/",root_node,sep=""), n_randsets, directory, root_node, "classic")
				}	
#				stop("only randomset")
				# category test
				hyper_category_test(paste(directory, "/randset_out",sep=""), paste(directory,"/category_test_out", sep=""), 1, root_node)
			} else if (test=="wilcoxon"){
				wilcox_randset(paste(directory,"/",root_node,sep=""), n_randsets, directory, root_node)
				# category test
				wilcox_category_test(paste(directory, "/randset_out",sep=""), paste(directory,"/category_test_out", sep=""), 1, root_node)
			}
		
			## 3c) summarize func-output
			groupy = read.table(paste(directory,"/category_test_out",sep=""))
			# NEW remove expected and actual no. of candidate, ranksum (included in c++ files to be the same for go_enrich)
			groupy = groupy[,1:5]
			
#			# check that FWER order follows p-value order
#			colnames(groupy)=c("node_id","p_under","p_over","FWER_under","FWER_over")
#			groupy_sorted = groupy[order(signif(groupy$p_over,12), -groupy$FWER_over),]
#			if(any(groupy_sorted$FWER_over != cummax(groupy_sorted$FWER_over))){
#				print(data.frame(groupy_sorted[,c(1,3,5)], FWER_check=groupy_sorted$FWER_over == cummax(groupy_sorted$FWER_over)))
#				stop("FWER_over does not strictly follow p_over. This looks like a bug.\n  Please contact steffi_grote@eva.mpg.de.")
#			}
#			groupy_sorted = groupy[order(signif(groupy$p_under,12), -groupy$FWER_under),]
#			if(any(groupy_sorted$FWER_under != cummax(groupy_sorted$FWER_under))){
#				print(data.frame(groupy_sorted[,c(1,2,4)], FWER_check=groupy_sorted$FWER_under == cummax(groupy_sorted$FWER_under)))
#				stop("FWER_under does not strictly follow p_under.")
#			}
			
			age_category = rep(s,nrow(groupy))			
			cutoff_quantile = rep(cutoff_quantiles[i],nrow(groupy))
			cutoff_value = rep(cutoff[i],nrow(groupy))
			groups = cbind(age_category,cutoff_quantile,cutoff_value,groupy)
				
			# combine with output from previous cutoffs
			out = rbind(out,groups)			
		} # end cutoffs		
	} # end ages
	 
	#####. 4 rearrange output
	message("Creating output...")
	
	if (test == "hyper"){
		colnames(out)[4:8] = c("node_id","raw_p_underrep","raw_p_overrep","FWER_underrep","FWER_overrep")
	} else if (test  == "wilcoxon"){
		colnames(out)[4:8] = c("node_id","raw_p_low_rank","raw_p_high_rank","FWER_low_rank","FWER_high_rank")
	}
	# remove redundant rows (nodes with no data and only one child)
	cluster = get(paste("node_clusters",folder_ext,sep="_"))
	results = rearrange_output(out,cluster,term)
	# rearrange cutoff-list
	cutoffs = do.call(cbind.data.frame, cutoff_list)
	colnames(cutoffs) = paste("age_category",colnames(cutoffs),sep="_")
	# sort genes alphabetically (useful for regions input)
	remaining_genes = remaining_genes[mixedorder(names(remaining_genes))]
	
	# save in package environment: dataframes for test and background-gene expression, test, dataset	
	# (to be returned with get_expression(structure_ids))	
	requested_gene_expression = pre_input
	rownames(requested_gene_expression) = NULL
	colnames(requested_gene_expression)[3] = "structure_id"
#	requested_gene_expression[,3] = paste("Allen:",requested_gene_expression[,3],sep="")
	remember = list(rge=requested_gene_expression, test=c(test,dataset), genes=remaining_genes)
	# save to package environment
	aba_env = as.environment("package:ABAEnrichment")
	unlock_environment(aba_env)
	if (exists("remember",where=aba_env)){
		unlockBinding("remember",aba_env)
	}	
	aba_env$remember = remember
	lockEnvironment(aba_env, bindings=TRUE)

	final_output = list(results=results, genes=remaining_genes, cutoffs=cutoffs)
	
	message("\nDone.\n")	
	return(final_output)
}

	
